/*
 * Copyright (c) 2008 Jiri Svoboda
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @addtogroup debug
 * @{
 */
/** @file
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <udebug.h>

#include "../../../cons.h"
#include "../../../main.h"
#include "../../../breakpoint.h"
#include "../../../include/arch.h"

#define OPCODE_BREAK		0x0000000d

static istate_t istate;

int arch_breakpoint_set(breakpoint_t *b)
{
	uint32_t brkp;
	int rc;

	rc = udebug_mem_read(app_phone, &b->arch.back, b->addr,
	    sizeof(b->arch.back));
	if (rc < 0) return rc;

	cons_printf("udebug_mem_read() -> %d\n", rc);
	brkp = OPCODE_BREAK;
	rc = udebug_mem_write(app_phone, &brkp, b->addr, sizeof(brkp));
	cons_printf("udebug_mem_write() -> %d\n", rc);
	if (rc < 0) return rc;

	return 0;
}

int arch_breakpoint_remove(breakpoint_t *b)
{
	int rc;

	if (b->active) {
		rc = udebug_mem_write(app_phone, &b->arch.back, b->addr + 4, 4);
		if (rc < 0) {
			cons_printf("error writing memory\n");
			return rc;
		}
		active_bkpt = NULL;
	} else {
	        rc = udebug_mem_write(app_phone, &b->arch.back, b->addr, 4);
		if (rc < 0) {
			cons_printf("error writing memory\n");
			return rc;
		}
	}

	return 0;

}

void arch_event_breakpoint(thash_t thread_hash)
{
	breakpoint_t *b;
	dthread_t *dt;
	int rc;
	uint32_t epc;
	int brk_addr;
	uint32_t brkp;

	brkp = OPCODE_BREAK;

	cons_printf("arch_event_breakpoint\n");

	rc = udebug_regs_read(app_phone, thread_hash, &istate);
	cons_printf("udebug_regs_read -> %d\n", rc);
	epc = istate_get_pc(&istate);
	cons_printf("EPC was 0x%08x\n", epc);
	brk_addr = epc;

	b = breakpoint_find_by_addr(brk_addr);
	if (b != NULL) {
		cons_printf("move breakpoint\b");
	        rc = udebug_mem_write(app_phone, &b->arch.back, brk_addr, 4);
		rc = udebug_mem_read(app_phone, &b->arch.back, brk_addr + 4, 4);
	        rc = udebug_mem_write(app_phone, &brkp, brk_addr + 4, 4);
		active_bkpt = b;
		b->active = true;

		cons_printf("breakpoint_hit...\n");
		breakpoint_hit(b);
		cons_printf("end_hit...\n");
		return;
	}

	b = breakpoint_find_by_addr(brk_addr - 4);
	if (b != NULL && b->active) {
		cons_printf("restoring breakpoint %d\n", b->id);
	        rc = udebug_mem_write(app_phone, &b->arch.back, brk_addr, 4);
		rc = udebug_mem_read(app_phone, &b->arch.back, brk_addr - 4, 4);
	        rc = udebug_mem_write(app_phone, &brkp, brk_addr - 4, 4);
		active_bkpt = NULL;
		
		if (dt->arch.singlestep) {
			singlestep_hit();

			rc = udebug_mem_read(app_phone, &dt->arch.sstep_back, brk_addr + 4, 4);
			rc = udebug_mem_write(app_phone, &brkp, brk_addr + 4, 4);
		}
		return;	
	}

	dt = dthread_get();

	if (dt->arch.singlestep) {
		cons_printf("advance singlestep\n");
	        rc = udebug_mem_write(app_phone, &dt->arch.sstep_back, brk_addr, 4);
		rc = udebug_mem_read(app_phone, &dt->arch.sstep_back, brk_addr + 4, 4);
	        rc = udebug_mem_write(app_phone, &brkp, brk_addr + 4, 4);

		singlestep_hit();

		return;
	}

	cons_printf("Unrecognized breakpoint at 0x%lx\n", brk_addr);
}

void arch_event_trap(dthread_t *dt)
{
	/* Unused */
	(void)dt;
}

void arch_dump_regs(thash_t thash)
{
}

void arch_set_singlestep(dthread_t *dt, int enable)
{
	int rc;
	uint32_t epc;
	uint32_t brk;
	breakpoint_t *b1, *b2;

	brk = OPCODE_BREAK;

	cons_printf("arch_set_singlestep(dt, %d)\n", enable);
	rc = udebug_regs_read(app_phone, dt->hash, &istate);
	cons_printf("udebug_regs_read -> %d\n", rc);
	epc = istate_get_pc(&istate);
	cons_printf("EPC was 0x%08x\n", epc);

	b1 = breakpoint_find_by_addr(epc - 4);
	b2 = breakpoint_find_by_addr(epc);

	if (enable && !dt->arch.singlestep) {
		if (b1 && b1->active) {
			dt->arch.sstep_back = b1->arch.back;
		} else if (b2) {
			dt->arch.sstep_back = b2->arch.back;
		} else {
			cons_printf("initial set singlestep\b");
			rc = udebug_mem_read(app_phone, &dt->arch.sstep_back, epc + 4, 4);
			rc = udebug_mem_write(app_phone, &brk, epc + 4, 4);
			if (rc < 0) { cons_printf("error writing mem\n"); return; }
		}
	} else if (!enable && dt->arch.singlestep) {
		if ((b1 && b1->active) || b2) {
			/* Do not remove BRK instruction */
		} else  {
			cons_printf("remove singlestep\b");
			rc = udebug_mem_write(app_phone, &dt->arch.sstep_back, epc + 4, 4);
		}
	}
	dt->arch.singlestep = enable;
}

/** @}
 */
